/* bmexample.c
 * BLINKENmini program example
 * Copyright (C) 2002 sphaera & 1stein (http://blinkenmini.1stein.no-ip.com/)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>

#define VerMaj 1
#define VerMin 0
#define VerRev 0

#define DisplaySizeX 18
#define DisplaySizeY 8

#define count( array ) (sizeof( (array) ) / sizeof( (array)[0] ))

//keys of controllers
#define KeyB 0
#define KeyY 1
#define KeySelect 2
#define KeyStart 3
#define KeyUp 4
#define KeyDown 5
#define KeyLeft 6
#define KeyRight 7
#define KeyA 8
#define KeyX 9
#define KeyL 10
#define KeyR 11

//current picture
unsigned char Picture[DisplaySizeY][DisplaySizeX];

//cursor
int CursorStep = 0, CursorX = (DisplaySizeX - 1) / 2, CursorY = (DisplaySizeY - 1) / 2;

//names of buttons
char * ButtonNames[] = { "B", "Y", "Select", "Start", "Up", "Down", "Left", "Right", "A", "X", "L", "R" };

//controller button was pressed or released
void ControllerAction( unsigned char Controller, unsigned char Button, unsigned char Pressed )
{
	//display message
	if( Button < count( ButtonNames ) )
		if( Pressed )
			printf( "button %s of controller %d was pressed\n", ButtonNames[Button], Controller );
		else
			printf( "button %s of controller %d was released\n", ButtonNames[Button], Controller );
	else
		if( Pressed )
			printf( "button %d of controller %d was pressed\n", Button, Controller );
		else
			printf( "button %d of controller %d was released\n", Button, Controller );

	//button was released
	if( !Pressed )
		return;

	//process button
	switch( Button )
	{
		//move cursor
		case KeyUp:
			CursorY--;
			if( CursorY < 0 )
				CursorY = DisplaySizeY - 1;
			break;
		case KeyDown:
			CursorY++;
			if( CursorY > DisplaySizeY - 1 )
				CursorY = 0;
			break;
		case KeyLeft:
			CursorX--;
			if( CursorX < 0 )
				CursorX = DisplaySizeX - 1;
			break;
		case KeyRight:
			CursorX++;
			if( CursorX > DisplaySizeX - 1 )
				CursorX = 0;
			break;
		//increase brightness of pixel
		case KeyB:
			if( Picture[CursorY][CursorX] < 0xF0 )
				Picture[CursorY][CursorX] += 0x10;
			else
				Picture[CursorY][CursorX] = 0xFF;
			break;
		//decrease brightness of pixel
		case KeyY:
			if( Picture[CursorY][CursorX] > 0x0F )
				Picture[CursorY][CursorX] -= 0x10;
			else
				Picture[CursorY][CursorX] = 0x00;
			break;
	}
}

//main
int main( int ArgC, char * * ArgV )
{
	int BmdrvOut, BmdrvIn, ReadBufferCnt = 0, I;
	unsigned char ReadBuffer[4096], FrameBuffer[DisplaySizeX * DisplaySizeY + 3];

	//print message
	printf( "\n" );
	printf( "bmexample: BLINKENmini program example (version %d.%d.%d)\n", VerMaj, VerMin, VerRev );
	printf( "           done in 2002 by sphaera & 1stein\n" );
	printf( "           http://blinkenmini.1stein.no-ip.com/\n" );
	printf( "           distributed under GNU Public License - no warranty\n\n" );
	//check arguments
	if( ArgC != 3 )
	{
		printf( "bmexample <out-bmdrv> <in-bmdrv>\n\n" );
		return 1;
	}

	//open bmdrv
	BmdrvOut = open( ArgV[1], O_WRONLY | O_NOCTTY | O_NONBLOCK );
	if( BmdrvOut == -1 )
	{
		printf( "bmexample: could not open %s: %s\n\n", ArgV[1], strerror( errno ) );
		return -1;
	}
	BmdrvIn = open( ArgV[2], O_RDONLY | O_NOCTTY | O_NONBLOCK );
	if( BmdrvIn == -1 )
	{
		printf( "bmexample: could not open %s: %s\n\n", ArgV[2], strerror( errno ) );
		close( BmdrvOut );
		return -1;
	}

	//clear current picture
	memset( Picture, 0, DisplaySizeX * DisplaySizeY );
	
	for( ; ; )
	{
		//read from bmdrv
		I = read( BmdrvIn, &ReadBuffer + ReadBufferCnt, sizeof( ReadBuffer ) - ReadBufferCnt );
		if( I == -1 && errno != EAGAIN )
		{
			printf( "bmexample: could not read from %s: %s\n\n", ArgV[2], strerror( errno ) );
			close( BmdrvIn );
			close( BmdrvOut );
			return -1;
		}
		if( I > 0 )
			ReadBufferCnt += I;

		//process controller button actions
		for( I = 0; I < ReadBufferCnt && ReadBuffer[I] != 0xFF; I++ ); //ignore everything before 0xFF
		for( ; I + 3 < ReadBufferCnt; I += 4 ) //as long, as complete actons are availible
			ControllerAction( ReadBuffer[I + 2], ReadBuffer[I + 3], ReadBuffer[I + 1] ); //process this action
		ReadBufferCnt -= I; //remove processed actions from buffer
		memmove( ReadBuffer, ReadBuffer + I, ReadBufferCnt );

		//copy current picture to frame buffer
		FrameBuffer[0] = DisplaySizeX;
		FrameBuffer[1] = DisplaySizeY;
		memcpy( FrameBuffer + 2, Picture, DisplaySizeX * DisplaySizeY );
		FrameBuffer[DisplaySizeX * DisplaySizeY + 2] = 0;

		//put cursor into frame buffer
		switch( CursorStep )
		{
			case 0: FrameBuffer[CursorY * DisplaySizeX + CursorX + 2] = 0x00; break;
			case 2: FrameBuffer[CursorY * DisplaySizeX + CursorX + 2] = 0xFF; break;
		}
		CursorStep++;
		if( CursorStep > 3 )
			CursorStep = 0;

		//write frame to bmdrv
		I = write( BmdrvOut, &FrameBuffer, DisplaySizeX * DisplaySizeY + 3 );
		if( I != DisplaySizeX * DisplaySizeY + 3 )
		{
			printf( "bmexample: could not write to %s: %s\n\n", ArgV[1], strerror( errno ) );
			close( BmdrvIn );
			close( BmdrvOut );
			return -1;
		}

		//wait 0.25 seconds
		usleep( 250000 );
	} //for( ; ; )

	//close bmdrv
	close( BmdrvIn );
	close( BmdrvOut );

	printf( "done...\n\n" );
	return 0;
}

